from itertools import groupby

from prefect import task


@task
def group_template_conditions(conditions, max_seeds):
    grouped_conditions = group_seed_ranges(conditions, max_seeds)
    condition_sets = group_condition_sets(grouped_conditions, max_seeds)
    template_conditions = [{"conditions": condition_set} for condition_set in condition_sets]
    return template_conditions


def group_seed_ranges(conditions, max_seeds):
    conditions.sort(key=lambda x: (x["key"], x["seed"]))
    grouped_conditions = []

    for key, condition_group in groupby(conditions, lambda x: x["key"]):
        key_seeds = []
        key_conditions = {}

        for condition in condition_group:
            key_seeds.append(condition["seed"])
            key_conditions.update(condition)

        key_conditions.pop("seed")
        seed_ranges = find_seed_ranges(key_seeds, max_seeds)

        for range_start, range_end in seed_ranges:
            group_condition = key_conditions.copy()
            group_condition["start_seed"] = range_start
            group_condition["end_seed"] = range_end
            grouped_conditions.append(group_condition)

    return grouped_conditions


def find_seed_ranges(seeds, max_seeds):
    seeds.sort()
    ranges = []

    for _, group in groupby(enumerate(seeds), lambda x: x[0] - x[1]):
        group_list = list(group)
        subset = True
        range_start = group_list[0][1]
        range_end = group_list[-1][1]

        while subset:
            ranges.append((range_start, min(range_end, range_start + max_seeds - 1)))
            range_start = range_start + max_seeds
            if range_start > range_end:
                subset = False

    return ranges


def group_condition_sets(conditions, max_seeds):
    seed_count = 0
    condition_set = []
    condition_sets = []

    for condition in conditions:
        num_seeds = condition["end_seed"] - condition["start_seed"] + 1

        if seed_count + num_seeds <= max_seeds:
            condition_set.append(condition)
            seed_count = seed_count + num_seeds
        else:
            condition_sets.append(condition_set)
            seed_count = num_seeds
            condition_set = [condition]

    condition_sets.append(condition_set)

    return condition_sets



import pandas as pd
from prefect import task


@task
def merge_parsed_results(*results: pd.DataFrame) -> pd.DataFrame:
    for result in results:
        result.set_index(["ID", "TICK"], inplace=True)

    merged_results = pd.concat(results, axis=1)

    return merged_results.reset_index()



from typing import List, Union
import json
import tarfile

import numpy as np
import pandas as pd
from prefect import task

CELLS_COLUMNS = [
    "ID",
    "TICK",
    "CENTER_X",
    "CENTER_Y",
    "CENTER_Z",
    "MIN_X",
    "MIN_Y",
    "MIN_Z",
    "MAX_X",
    "MAX_Y",
    "MAX_Z",
]


@task
def parse_locations_file(tar: tarfile.TarFile, regions: list[str]) -> pd.DataFrame:
    all_locations: List[List[Union[str, int]]] = []

    for member in tar.getmembers():
        timepoint = int(member.name.replace(".LOCATIONS.json", "").split("_")[-1])

        extracted_member = tar.extractfile(member)
        assert extracted_member is not None
        locations_json = json.loads(extracted_member.read().decode("utf-8"))

        locations = [parse_location_timepoint(timepoint, cell, regions) for cell in locations_json]
        all_locations = all_locations + locations

    columns = CELLS_COLUMNS + [
        f"{column}.{region}" for region in regions for column in CELLS_COLUMNS[2:]
    ]
    locations_df = pd.DataFrame(all_locations, columns=columns)

    return locations_df


def parse_location_timepoint(timepoint: int, loc: dict, regions: list[str]) -> list:
    if "center" in loc:
        voxels = np.array([voxel for region in loc["location"] for voxel in region["voxels"]])
        mins = np.min(voxels, axis=0)
        maxs = np.max(voxels, axis=0)
        parsed = [loc["id"], timepoint, *loc["center"], *mins, *maxs]
    else:
        parsed = [loc["id"], timepoint, -1, -1, -1, -1, -1, -1, -1, -1, -1]

    for reg in regions:
        region_voxels = np.array(
            [
                voxel
                for region in loc["location"]
                for voxel in region["voxels"]
                if region["region"] == reg
            ]
        )

        if len(region_voxels) == 0:
            parsed = parsed + [-1, -1, -1, -1, -1, -1, -1, -1, -1]
            continue

        center = [round(value) for value in region_voxels.mean(axis=0)]
        mins = np.min(region_voxels, axis=0)
        maxs = np.max(region_voxels, axis=0)
        parsed = parsed + [*center, *mins, *maxs]

    return parsed



from typing import List, Union
import json
import tarfile

import pandas as pd
from prefect import task

CELLS_COLUMNS = [
    "ID",
    "TICK",
    "PARENT",
    "POPULATION",
    "AGE",
    "DIVISIONS",
    "STATE",
    "PHASE",
    "NUM_VOXELS",
]


@task
def parse_cells_file(tar: tarfile.TarFile, regions: list[str]) -> pd.DataFrame:
    all_cells: List[List[Union[str, int]]] = []

    for member in tar.getmembers():
        timepoint = int(member.name.replace(".CELLS.json", "").split("_")[-1])

        extracted_member = tar.extractfile(member)
        assert extracted_member is not None
        cells_json = json.loads(extracted_member.read().decode("utf-8"))

        cells = [parse_cell_timepoint(timepoint, cell, regions) for cell in cells_json]
        all_cells = all_cells + cells

    columns = CELLS_COLUMNS + [f"NUM_VOXELS.{region}" for region in regions]
    cells_df = pd.DataFrame(all_cells, columns=columns)

    return cells_df


def parse_cell_timepoint(timepoint: int, cell: dict, regions: list[str]) -> list:
    features = ["parent", "pop", "age", "divisions", "state", "phase", "voxels"]
    parsed = [cell["id"], timepoint] + [cell[feature] for feature in features]

    if regions and "regions" in cell:
        region_voxels = [
            cell_region["voxels"]
            for region in regions
            for cell_region in cell["regions"]
            if cell_region["region"] == region
        ]
        parsed = parsed + region_voxels

    return parsed
